@@ -30,6 +30,12 @@ let try_with_result fn a =
   try Ok (fn a)
   with Parse_error (msg, _) -> Error (`Msg ("Ipaddr: " ^ msg))
 
+let failwith_msg = function
+  | Ok x -> x
+  | Error (`Msg m) -> failwith m
+
+let map_result v f = match v with Ok v -> Ok (f v) | Error _ as e -> e
+
 let string_of_scope = function
 | Point -> "point"
 | Interface -> "interface"
@@ -68,9 +74,9 @@ let char_a = int_of_char 'a'
 let char_A = int_of_char 'A'
 
 let int_of_char c = match c with
-  | '0'..'9' -> Pervasives.int_of_char c - char_0
-  | 'a'..'f' -> 10 + Pervasives.int_of_char c - char_a
-  | 'A'..'F' -> 10 + Pervasives.int_of_char c - char_A
+  | '0'..'9' -> Stdlib.int_of_char c - char_0
+  | 'a'..'f' -> 10 + Stdlib.int_of_char c - char_a
+  | 'A'..'F' -> 10 + Stdlib.int_of_char c - char_A
   | _ -> -1
 
 let bad_char i s =
@@ -266,6 +272,18 @@ module V4 = struct
       end
     | _ -> None
 
+  let succ t =
+    if Int32.equal t 0xFF_FF_FF_FFl then
+      Error (`Msg "Ipaddr: highest address has been reached")
+    else
+      Ok (Int32.succ t)
+
+  let pred t =
+    if Int32.equal t 0x00_00_00_00l then
+      Error (`Msg "Ipaddr: lowest address has been reached")
+    else
+      Ok (Int32.pred t)
+
   (* constant *)
 
   let any         = make   0   0   0   0
@@ -281,7 +299,7 @@ module V4 = struct
 
     let compare (pre,sz) (pre',sz') =
       let c = compare pre pre' in
-      if c = 0 then Pervasives.compare sz sz' else c
+      if c = 0 then Stdlib.compare sz sz' else c
 
     let ip = make
 
@@ -290,10 +308,12 @@ module V4 = struct
       else if sz >= 32 then 0x0_FF_FF_FF_FF_l
       else 0x0_FF_FF_FF_FF_l <|< (32 - sz)
 
-    let make sz pre = (pre &&& (mask sz),sz)
+    let prefix (pre,sz) = (pre &&& (mask sz), sz)
+
+    let make sz pre = (pre,sz)
 
     let network_address (pre,sz) addr =
-      pre ||| (addr &&& Int32.lognot (mask sz))
+      (pre &&& (mask sz)) ||| (addr &&& Int32.lognot (mask sz))
 
     (* string conversion *)
 
@@ -319,20 +339,20 @@ module V4 = struct
 
     let of_string s = try_with_result of_string_exn s
 
-    let of_address_string_exn s =
-      let (p,quad) = _of_string_exn s in (make p quad, quad)
-
-    let of_address_string s = try_with_result of_address_string_exn s
-
-    let of_netmask nm addr =
+    let _of_netmask_exn ~netmask address =
       let rec find_greatest_one bits i =
         if bits = 0_l then i-1 else find_greatest_one (bits >|> 1) (i+1)
       in
-      let one = nm &&& (Int32.neg nm) in
+      let one = netmask &&& (Int32.neg netmask) in
       let sz = 32 - (find_greatest_one one (if one = 0_l then 33 else 0)) in
-      if nm <> (mask sz)
-      then raise (Parse_error ("invalid netmask",to_string nm))
-      else make sz addr
+      if netmask <> (mask sz)
+      then raise (Parse_error ("invalid netmask",to_string netmask))
+      else make sz address
+
+    let of_netmask_exn ~netmask ~address = _of_netmask_exn ~netmask address
+
+    let of_netmask ~netmask ~address =
+      try_with_result (_of_netmask_exn ~netmask) address
 
     let to_buffer buf (pre,sz) = Printf.bprintf buf "%a/%d" to_buffer pre sz
 
@@ -344,15 +364,9 @@ module V4 = struct
     let pp ppf i =
       Format.fprintf ppf "%s" (to_string i)
 
-    let to_address_buffer buf ((_,sz) as subnet) addr =
-      to_buffer buf (network_address subnet addr,sz)
-
-    let to_address_string subnet addr =
-      let b = Buffer.create 18 in
-      to_address_buffer b subnet addr;
-      Buffer.contents b
-
-    let mem ip (pre,sz) = let host = 32 - sz in (ip >|> host) = (pre >|> host)
+    let mem ip (pre,sz) =
+      let m = mask sz in
+      (ip &&& m) = (pre &&& m)
 
     let subset ~subnet:(pre1,sz1) ~network:(pre2,sz2) =
       sz1 >= sz2 && mem pre1 (pre2,sz2)
@@ -378,9 +392,22 @@ module V4 = struct
     ]
 
     let broadcast (pre,sz) = pre ||| (0x0_FF_FF_FF_FF_l >|> sz)
-    let network (pre,_) = pre
+    let network (pre,sz) = pre &&& mask sz
+    let address (addr,_) = addr
     let bits (_,sz) = sz
     let netmask subnet = mask (bits subnet)
+
+    let first (_,sz as cidr) =
+      if sz > 30 then
+        network cidr
+      else
+        network cidr |> succ |> failwith_msg
+
+    let last (_,sz as cidr) =
+      if sz > 30 then
+        broadcast cidr
+      else
+        broadcast cidr |> pred |> failwith_msg
   end
 
   (* TODO: this could be optimized with something trie-like *)
@@ -459,6 +486,59 @@ module B128 = struct
     (a1 ||| a2, b1 ||| b2, c1 ||| c2, d1 ||| d2)
 
   let lognot (a,b,c,d) = Int32.(lognot a, lognot b, lognot c, lognot d)
+
+  let succ (a,b,c,d) =
+    let cb (n,tl) v =
+      match n with
+      | 0l -> (0l,v::tl)
+      | n ->
+         let n =
+           if Int32.equal v 0xFF_FF_FF_FFl then
+             n
+           else
+             0l
+         in
+         (n,Int32.succ v::tl)
+    in
+    match List.fold_left cb (1l,[]) [d;c;b;a] with
+    | 0l, [a;b;c;d] -> Ok (of_int32 (a,b,c,d))
+    | n, [_;_;_;_] when n > 0l ->
+      Error (`Msg "Ipaddr: highest address has been reached")
+    | _ -> Error (`Msg "Ipaddr: unexpected error with B128")
+
+  let pred (a,b,c,d) =
+    let cb (n,tl) v =
+      match n with
+      | 0l -> (0l,v::tl)
+      | n ->
+         let n =
+           if v = 0x00_00_00_00l then
+             n
+           else
+             0l
+         in
+         (n,Int32.pred v::tl)
+    in
+    match List.fold_left cb (-1l,[]) [d;c;b;a] with
+    | 0l, [a;b;c;d] -> Ok (of_int32 (a,b,c,d))
+    | n, [_;_;_;_] when n < 0l ->
+      Error (`Msg "Ipaddr: lowest address has been reached")
+    | _ -> Error (`Msg "Ipaddr: unexpected error with B128")
+
+  let shift_right (a,b,c,d) sz =
+    let rec loop (a,b,c,d) sz =
+      if sz < 32 then (sz, (a,b,c,d))
+      else loop (0l,a,b,c) (sz - 32)
+    in
+    let (sz, (a,b,c,d)) = loop (a,b,c,d) sz in
+    let fn (saved,tl) part =
+      let new_saved = Int32.logand part (0xFF_FF_FF_FFl >|> sz) in
+      let new_part = (part >|> sz) ||| (saved <|< 32 - sz) in
+      (new_saved, new_part::tl)
+    in
+    match List.fold_left fn (0l,[]) [a;b;c;d] with
+    | _, [d;c;b;a] -> Ok (of_int32 (a, b, c, d))
+    | _ -> Error (`Msg "Ipaddr: unexpected error with B128.shift_right")
 end
 
 module V6 = struct
@@ -750,7 +830,7 @@ module V6 = struct
 
     let compare (pre,sz) (pre',sz') =
       let c = compare pre pre' in
-      if c = 0 then Pervasives.compare sz sz' else c
+      if c = 0 then Stdlib.compare sz sz' else c
 
     let ip = make
 
@@ -764,10 +844,12 @@ module V6 = struct
       mask (sz - 64),
       mask (sz - 96))
 
-    let make sz pre = (logand pre (mask sz),sz)
+    let prefix (pre,sz) = (logand pre (mask sz),sz)
+
+    let make sz pre = (pre,sz)
 
     let network_address (pre,sz) addr =
-      logor pre (logand addr (lognot (mask sz)))
+      logor (logand pre (mask sz)) (logand addr (lognot (mask sz)))
 
     let _of_string_raw s i =
       let v6 = of_string_raw s i in
@@ -791,20 +873,25 @@ module V6 = struct
 
     let of_string s = try_with_result of_string_exn s
 
-    let of_address_string_exn s =
-      let (p,v6) = _of_string_exn s in (make p v6, v6)
+    let _of_netmask_exn ~netmask address =
+      let nm =
+        let bits netmask =
+          V4.Prefix.bits (V4.Prefix.of_netmask_exn ~netmask ~address:V4.any)
+        in
+        match netmask with
+        | (0_l,0_l,0_l,0_l) -> 0
+        | (lsw ,0_l ,0_l ,0_l) -> bits lsw
+        | (-1_l,lsw ,0_l ,0_l) -> bits lsw + 32
+        | (-1_l,-1_l,lsw ,0_l) -> bits lsw + 64
+        | (-1_l,-1_l,-1_l,lsw) -> bits lsw + 96
+        | _ -> raise (Parse_error ("invalid netmask", to_string netmask))
+      in
+      make nm address
 
-    let of_address_string s = try_with_result of_address_string_exn s
+    let of_netmask_exn ~netmask ~address = _of_netmask_exn ~netmask address
 
-    let of_netmask nm addr =
-      make (match nm with
-      | (0_l,0_l,0_l,0_l) -> 0
-      | (lsw ,0_l ,0_l ,0_l) -> V4.Prefix.(bits (of_netmask lsw V4.any))
-      | (-1_l,lsw ,0_l ,0_l) -> V4.Prefix.(bits (of_netmask lsw V4.any)) + 32
-      | (-1_l,-1_l,lsw ,0_l) -> V4.Prefix.(bits (of_netmask lsw V4.any)) + 64
-      | (-1_l,-1_l,-1_l,lsw) -> V4.Prefix.(bits (of_netmask lsw V4.any)) + 96
-      | _ -> raise (Parse_error ("invalid netmask", to_string nm))
-      ) addr
+    let of_netmask ~netmask ~address =
+      try_with_result (_of_netmask_exn ~netmask) address
 
     let to_buffer buf (pre,sz) =
       Printf.bprintf buf "%a/%d" to_buffer pre sz
@@ -817,14 +904,6 @@ module V6 = struct
     let pp ppf i =
       Format.fprintf ppf "%s" (to_string i)
 
-    let to_address_buffer buf ((_,sz) as subnet) addr =
-      to_buffer buf (network_address subnet addr,sz)
-
-    let to_address_string subnet addr =
-      let b = Buffer.create 43 in
-      to_address_buffer b subnet addr;
-      Buffer.contents b
-
     let mem ip (pre,sz) =
       let m = mask sz in
       logand ip m = logand pre m
@@ -842,9 +921,21 @@ module V6 = struct
     let noneui64_interface  = make   3 (ip 0x0000 0 0 0 0 0 0 0)
     let solicited_node      = make 104 (ip 0xff02 0 0 0 0 1 0xff00 0)
 
-    let network (pre,_) = pre
+    let network (pre,sz) = logand pre (mask sz)
+    let address (addr,_) = addr
     let bits (_,sz) = sz
     let netmask subnet = mask (bits subnet)
+
+    let first (_,sz as cidr) =
+      if sz > 126 then
+        network cidr
+      else
+        network cidr |> succ |> failwith_msg
+
+    let last (_,sz as cidr) =
+      let ffff = ip 0xffff 0xffff 0xffff 0xffff
+          0xffff 0xffff 0xffff 0xffff in
+      logor (network cidr) (shift_right ffff sz |> failwith_msg)
   end
 
   (* TODO: This could be optimized with something trie-like *)
@@ -978,6 +1069,14 @@ let of_domain_name n =
     end
   | _ -> None
 
+let succ = function
+  | V4 addr -> map_result (V4.succ addr) (fun v -> V4 v)
+  | V6 addr -> map_result (V6.succ addr) (fun v -> V6 v)
+
+let pred = function
+  | V4 addr -> map_result (V4.pred addr) (fun v -> V4 v)
+  | V6 addr -> map_result (V6.pred addr) (fun v -> V6 v)
+
 module Prefix = struct
   module Addr = struct
     let to_v6 = to_v6
@@ -1052,4 +1151,13 @@ module Prefix = struct
 
   let pp ppf i =
     Format.fprintf ppf "%s" (to_string i)
+
+  let first = function
+    | V4 p -> V4 (V4.Prefix.first p)
+    | V6 p -> V6 (V6.Prefix.first p)
+
+  let last = function
+    | V4 p -> V4 (V4.Prefix.last p)
+    | V6 p -> V6 (V6.Prefix.last p)
+
 end
