@@ -148,6 +148,16 @@ module V4 : sig
       suffix, and an IPv4 address prefixed. *)
   val of_domain_name : 'a Domain_name.t -> t option
 
+  (** {3 Utility functions} *)
+
+  (** [succ ipv4] is ip address next to [ipv4].
+      Returns a human-readable error string if it's already the highest address. *)
+  val succ : t -> (t, [> `Msg of string ]) result
+
+  (** [pred ipv4] is ip address before [ipv4].
+      Returns a human-readable error string if it's already the lowest address. *)
+  val pred : t -> (t, [> `Msg of string ]) result
+
   (** {3 Common addresses} *)
 
   (** [any] is 0.0.0.0. *)
@@ -168,20 +178,24 @@ module V4 : sig
   (** [localhost] is 127.0.0.1. *)
   val localhost : t
 
-  (** A module for manipulating IPv4 network prefixes. *)
+  (** A module for manipulating IPv4 network prefixes (CIDR). *)
   module Prefix : sig
     type addr = t
 
-    (** Type of a internet protocol subnet *)
+    (** Type of a internet protocol subnet: an address and prefix length. *)
     type t
 
     (** [mask n] is the pseudo-address of an [n] bit subnet mask. *)
     val mask : int -> addr
 
-    (** [make n addr] is the [n] bit subnet prefix to which [addr] belongs. *)
+    (** [make n addr] is the cidr of [addr] with [n] bits prefix. *)
     val make : int -> addr -> t
 
-    (** [network_address prefix addr] is the address with prefix [prefix]
+    (** [prefix cidr] is the subnet prefix of [cidr] where all non-prefix bits
+        set to 0. *)
+    val prefix : t -> t
+
+    (** [network_address cidr addr] is the address with prefix [cidr]
         and suffix from [addr].
         See <http://tools.ietf.org/html/rfc4291#section-2.3>. *)
     val network_address : t -> addr -> addr
@@ -201,38 +215,26 @@ module V4 : sig
         into the string for reading. *)
     val of_string_raw : string -> int ref -> t
 
-    (** [to_string prefix] is the CIDR notation string representation
-        of [prefix], i.e. [XXX.XX.X.XXX/XX]. *)
+    (** [to_string cidr] is the CIDR notation string representation
+        of [cidr], i.e. [XXX.XX.X.XXX/XX]. *)
     val to_string : t -> string
 
-    (** [pp f prefix] outputs a human-readable representation of [prefix]
+    (** [pp f cidr] outputs a human-readable representation of [cidr]
         to the formatter [f]. *)
     val pp : Format.formatter -> t -> unit [@@ocaml.toplevel_printer]
 
-    (** [of_address_string_exn cidr_addr] is the address and prefix
-        represented by [cidr_addr]. Raises {!Parse_error} if [cidr_addr] is not
-        a valid representation of a CIDR-scoped address. *)
-    val of_address_string_exn : string -> t * addr
-
-    (** Same as {!of_address_string_exn} but returns a result type instead of
-        raising an exception. *)
-    val of_address_string : string -> (t * addr, [> `Msg of string ]) result
-
-    (** [to_address_string prefix addr] is the network address
-        constructed from [prefix] and [addr]. *)
-    val to_address_string : t -> addr -> string
-
-    (** [to_buffer buf prefix] writes the string representation
-        of [prefix] into the buffer [buf]. *)
+    (** [to_buffer buf cidr] writes the string representation
+        of [cidr] into the buffer [buf]. *)
     val to_buffer : Buffer.t -> t -> unit
 
-    (** [to_address_buffer buf prefix addr] writes string representation of the
-        network address representing [addr] in [prefix] to the buffer [buf]. *)
-    val to_address_buffer : Buffer.t -> t -> addr -> unit
+    (** [of_netmask_exn ~netmask ~address] is the subnet prefix of [address]
+        with netmask [netmask]. *)
+    val of_netmask_exn : netmask:addr -> address:addr -> t
 
-    (** [of_netmask netmask addr] is the subnet prefix of [addr] with netmask
-        [netmask]. *)
-    val of_netmask : addr -> addr -> t
+    (** [of_netmask ~netmask ~address] is the cidr of [address] with
+        netmask [netmask]. *)
+    val of_netmask : netmask:addr -> address:addr ->
+      (t, [> `Msg of string ]) result
 
     (** [mem ip subnet] checks whether [ip] is found within [subnet]. *)
     val mem : addr -> t -> bool
@@ -282,9 +284,18 @@ module V4 : sig
     (** [netmask subnet] is the netmask for [subnet]. *)
     val netmask : t -> addr
 
-    (** [bits subnet] is the bit size of the [subnet] prefix. *)
+    (** [address cidr] is the address for [cidr]. *)
+    val address : t -> addr
+
+    (** [bits cidr] is the bit size of the [cidr] prefix. *)
     val bits : t -> int
 
+    (** [first cidr] is first valid unicast address in this [cidr]. *)
+    val first : t -> addr
+
+    (** [last cidr] is last valid unicast address in this [cidr]. *)
+    val last : t -> addr
+
     include Map.OrderedType with type t := t
   end
 
@@ -365,7 +376,7 @@ module V6 : sig
       from offset [off] within [b].  [b] must be at least [off+16] bytes long or
       an error is returned. *)
   val write_octets : ?off:int -> t -> bytes -> (unit, [> `Msg of string ]) result
- 
+
   (** [to_octets ipv6] returns the 16 bytes representing the [ipv6] octets. *)
   val to_octets : t -> string
 
@@ -407,6 +418,16 @@ module V6 : sig
       suffix, and an IPv6 address prefixed. *)
   val of_domain_name : 'a Domain_name.t -> t option
 
+  (** {3 Utility functions} *)
+
+  (** [succ ipv6] is ip address next to [ipv6]. Returns a human-readable
+      error string if it's already the highest address. *)
+  val succ : t -> (t, [> `Msg of string ]) result
+
+  (** [pred ipv6] is ip address before [ipv6]. Returns a human-readable
+      error string if it's already the lowest address. *)
+  val pred : t -> (t, [> `Msg of string ]) result
+
   (** {3 Common addresses} *)
 
   (** [unspecified] is ::. *)
@@ -430,20 +451,24 @@ module V6 : sig
   (** [site_routers] is ff05::02. *)
   val site_routers : t
 
-  (** A module for manipulating IPv6 network prefixes. *)
+  (** A module for manipulating IPv6 network prefixes (CIDR). *)
   module Prefix : sig
     type addr = t
 
-    (** Type of a internet protocol subnet *)
+    (** Type of a internet protocol subnet: an address and a prefix length. *)
     type t
 
     (** [mask n] is the pseudo-address of an [n] bit subnet mask. *)
     val mask : int -> addr
 
-    (** [make n addr] is the [n] bit subnet prefix to which [addr] belongs. *)
+    (** [make n addr] is the cidr of [addr] with [n] bit prefix. *)
     val make : int -> addr -> t
 
-    (** [network_address prefix addr] is the address with prefix [prefix]
+    (** [prefix cidr] is the subnet prefix of [cidr] where all non-prefix bits
+        set to 0. *)
+    val prefix : t -> t
+
+    (** [network_address cidr addr] is the address with prefix [cidr]
         and suffix from [addr].
         See <http://tools.ietf.org/html/rfc4291#section-2.3>. *)
     val network_address : t -> addr -> addr
@@ -461,38 +486,26 @@ module V6 : sig
         into the string for reading. *)
     val of_string_raw : string -> int ref -> t
 
-    (** [to_string prefix] is the CIDR notation string representation
-        of [prefix], i.e. XXX:XX:X::XXX/XX. *)
+    (** [to_string cidr] is the CIDR notation string representation
+        of [cidr], i.e. XXX:XX:X::XXX/XX. *)
     val to_string : t -> string
 
-    (** [pp f prefix] outputs a human-readable representation of [prefix]
+    (** [pp f cidr] outputs a human-readable representation of [cidr]
         to the formatter [f]. *)
     val pp : Format.formatter -> t -> unit [@@ocaml.toplevel_printer]
 
-    (** [of_address_string_exn cidr_addr] is the address and prefix
-        represented by [cidr_addr]. Raises {!Parse_error} if [cidr_addr] is not
-        a valid representation of a CIDR-scoped address. *)
-    val of_address_string_exn : string -> t * addr
-
-    (** Same as {!of_address_string_exn} but returns an option type instead of
-        raising an exception. *)
-    val of_address_string : string -> ((t * addr), [> `Msg of string]) result
-
-    (** [to_address_string prefix addr] is the network address
-        constructed from [prefix] and [addr]. *)
-    val to_address_string : t -> addr -> string
-
-    (** [to_buffer buf prefix] writes the string representation
-        of [prefix] to the buffer [buf]. *)
+    (** [to_buffer buf cidr] writes the string representation
+        of [cidr] to the buffer [buf]. *)
     val to_buffer : Buffer.t -> t -> unit
 
-    (** [to_address_buffer buf prefix addr] writes string representation of the
-        network address representing [addr] in [prefix] to the buffer [buf]. *)
-    val to_address_buffer : Buffer.t -> t -> addr -> unit
+    (** [of_netmask_exn ~netmask ~address] is the subnet prefix of [address]
+        with netmask [netmask]. *)
+    val of_netmask_exn : netmask:addr -> address:addr -> t
 
-    (** [of_netmask netmask addr] is the subnet prefix of [addr] with netmask
-        [netmask]. *)
-    val of_netmask : addr -> addr -> t
+    (** [of_netmask ~netmask ~address] is the cidr of [address] with
+        netmask [netmask]. *)
+    val of_netmask : netmask:addr -> address:addr ->
+      (t, [> `Msg of string ]) result
 
     (** [mem ip subnet] checks whether [ip] is found within [subnet]. *)
     val mem : addr -> t -> bool
@@ -533,9 +546,18 @@ module V6 : sig
     (** [netmask subnet] is the netmask for [subnet]. *)
     val netmask : t -> addr
 
+    (** [address cidr] is the address for [cidr]. *)
+    val address : t -> addr
+
     (** [bits subnet] is the bit size of the [subnet] prefix. *)
     val bits : t -> int
 
+    (** [first subnet] is first valid unicast address in this [subnet]. *)
+    val first : t -> addr
+
+    (** [last subnet] is last valid unicast address in this [subnet]. *)
+    val last : t -> addr
+
     include Map.OrderedType with type t := t
   end
 
@@ -637,6 +659,14 @@ val to_domain_name : t -> [ `host ] Domain_name.t
     [ip6.arpa] suffix, and an IP address prefixed. *)
 val of_domain_name : 'a Domain_name.t -> t option
 
+(** [succ addr] is ip address next to [addr]. Returns a human-readable
+    error string if it's already the highest address. *)
+val succ : t -> (t, [> `Msg of string ]) result
+
+(** [pred addr] is ip address before [addr]. Returns a human-readable
+    error string if it's already the lowest address. *)
+val pred : t -> (t, [> `Msg of string ]) result
+
 module Prefix : sig
   type addr = t
 
@@ -696,6 +726,12 @@ module Prefix : sig
   (** [netmask subnet] is the netmask for [subnet]. *)
   val netmask : t -> addr
 
+  (** [first subnet] is first valid unicast address in this [subnet]. *)
+  val first : t -> addr
+
+  (** [last subnet] is last valid unicast address in this [subnet]. *)
+  val last : t -> addr
+
   include Map.OrderedType with type t := t
 end
 
