@@ -23,6 +23,8 @@ let need_more s = error s "not enough data"
 let bad_char i s =
   error s (Printf.sprintf "invalid character '%c' at %d" s.[i] i)
 
+let (>>=) v f = match v with Ok v -> f v | Error _ as e -> e
+
 let assert_raises ~msg exn test_fn =
   assert_raises ~msg exn (fun () ->
     try test_fn ()
@@ -112,7 +114,7 @@ module Test_v4 = struct
       "192.168.0.0/0",  "0.0.0.0/0";
     ] in
     List.iter (fun (subnet,rt) ->
-      let os = V4.Prefix.of_string_exn subnet in
+      let os = V4.Prefix.of_string_exn subnet |> V4.Prefix.prefix in
       let ts = V4.Prefix.to_string os in
       assert_equal ~msg:subnet ts rt;
       let os = Ipaddr_sexp.(V4.Prefix.(t_of_sexp (sexp_of_t os))) in
@@ -139,10 +141,13 @@ module Test_v4 = struct
     List.iter (fun (netaddr,net,addr) ->
       let netv4 = V4.Prefix.of_string_exn net in
       let addrv4 = V4.of_string_exn addr in
-      let prefix,v4 = V4.Prefix.of_address_string_exn netaddr in
+      let cidr = V4.Prefix.of_string_exn netaddr in
+      let prefix = V4.Prefix.prefix cidr
+      and v4 = V4.Prefix.address cidr
+      in
       assert_equal ~msg:(net^" <> "^(V4.Prefix.to_string prefix)) netv4 prefix;
       assert_equal ~msg:(addr^" <> "^(V4.to_string v4)) addrv4 v4;
-      let addrstr = V4.Prefix.to_address_string prefix v4 in
+      let addrstr = V4.Prefix.to_string cidr in
       assert_equal ~msg:(netaddr^" <> "^addrstr) netaddr addrstr;
     ) netaddrs
 
@@ -183,13 +188,16 @@ module Test_v4 = struct
       "192.168.0.1/0", "0.0.0.0";
     ] in
     List.iter (fun (net_str,nm_str) ->
-      let prefix, v4 = V4.Prefix.of_address_string_exn net_str in
-      let nm = V4.Prefix.netmask prefix in
-      let nnm_str = V4.to_string nm in
+      let cidr = V4.Prefix.of_string_exn net_str in
+      let prefix = V4.Prefix.prefix cidr
+      and address = V4.Prefix.address cidr
+      in
+      let netmask = V4.Prefix.netmask prefix in
+      let nnm_str = V4.to_string netmask in
       let msg = Printf.sprintf "netmask %s <> %s" nnm_str nm_str in
       assert_equal ~msg nnm_str nm_str;
-      let prefix = V4.Prefix.of_netmask nm v4 in
-      let nns = V4.Prefix.to_address_string prefix v4 in
+      let prefix = V4.Prefix.of_netmask_exn ~netmask ~address in
+      let nns = V4.Prefix.to_string prefix in
       let msg = Printf.sprintf "%s is %s under netmask iso" net_str nns in
       assert_equal ~msg net_str nns
     ) nets
@@ -200,9 +208,10 @@ module Test_v4 = struct
       error "255.255.254.128" "invalid netmask";
     ] in
     List.iter (fun (nm_str,exn) ->
-      let nm = V4.of_string_exn nm_str in
-      let addr = V4.of_string_exn "192.168.0.1" in
-      assert_raises ~msg:nm_str exn (fun () -> V4.Prefix.of_netmask nm addr)
+      let netmask = V4.of_string_exn nm_str in
+      let address = V4.of_string_exn "192.168.0.1" in
+      assert_raises ~msg:nm_str exn
+        (fun () -> V4.Prefix.of_netmask_exn ~netmask ~address)
     ) bad_masks
 
   let test_scope () =
@@ -252,11 +261,12 @@ module Test_v4 = struct
 
   let test_prefix_map () =
     let module M = Map.Make(V4.Prefix) in
-    let m = M.add (V4.Prefix.of_string_exn "0.0.0.0/0") "everyone" M.empty in
-    let m = M.add (V4.Prefix.of_string_exn "192.0.0.0/1") "weirdos" m in
-    let m = M.add (V4.Prefix.of_string_exn "128.0.0.0/1") "high-bitters" m in
-    let m = M.add (V4.Prefix.of_string_exn "254.0.0.0/8") "top-end" m in
-    let m = M.add (V4.Prefix.of_string_exn "0.0.0.0/0") "iana" m in
+    let of_string s = s |> V4.Prefix.of_string_exn |> V4.Prefix.prefix in
+    let m = M.add (of_string "0.0.0.0/0") "everyone" M.empty in
+    let m = M.add (of_string "192.0.0.0/1") "weirdos" m in
+    let m = M.add (of_string "128.0.0.0/1") "high-bitters" m in
+    let m = M.add (of_string "254.0.0.0/8") "top-end" m in
+    let m = M.add (of_string "0.0.0.0/0") "iana" m in
     assert_equal ~msg:"size" (M.cardinal m) 3;
     assert_equal ~msg:"min" (M.min_binding m)
       (V4.Prefix.of_string_exn "0.0.0.0/0", "iana");
@@ -304,6 +314,68 @@ module Test_v4 = struct
         (fun () -> Ipaddr_cstruct.V4.of_cstruct_exn (Cstruct.of_string addr))
     ) addrs
 
+  let test_prefix_mem () =
+    let ip = V4.of_string_exn in
+    let prefix = V4.Prefix.of_string_exn in
+    let ships = [
+      ip "10.0.0.7",            prefix "10.0.0.0/29",            true;
+      ip "172.16.255.254",      prefix "172.16.255.254/31",      true;
+      ip "192.168.0.1",         prefix "0.0.0.0/0",              true;
+      ip "192.168.0.1",         V4.Prefix.private_192,           true;
+      ip "255.255.255.255",     prefix "255.255.255.255/32",     true;
+      ip "192.0.2.1",           prefix "192.0.2.0/32",           false;
+      ip "192.0.2.1",           prefix "192.0.0.0/23",           false;
+      ip "255.255.255.255",     prefix "0.0.0.0/1",              false;
+    ] in
+    List.iter (fun (addr,subnet,is_mem) ->
+      let msg = Printf.sprintf "%s is%s in %s"
+        (V4.to_string addr) (if is_mem then "" else " not") (V4.Prefix.to_string subnet)
+      in
+      assert_equal ~msg (V4.Prefix.mem addr subnet) is_mem
+    ) ships
+
+  let test_succ_pred () =
+    let open V4 in
+    let printer = function
+      | Ok v -> Printf.sprintf "Ok %s" (to_string v)
+      | Error (`Msg e) -> Printf.sprintf "Error `Msg \"%s\"" e
+    in
+    let assert_equal = assert_equal ~printer in
+    let ip1 = of_string_exn "0.0.0.0" in
+    let ip2 = of_string_exn "255.255.255.255" in
+    assert_equal ~msg:"succ 0.0.0.0"
+      (of_string "0.0.0.1") (succ ip1);
+    assert_equal ~msg:"succ 255.255.255.255"
+      (Error (`Msg "Ipaddr: highest address has been reached")) (succ ip2);
+    assert_equal ~msg:"succ (succ 255.255.255.255)"
+      (Error (`Msg "Ipaddr: highest address has been reached"))
+      (succ ip2 >>= succ);
+    assert_equal ~msg:"pred 0.0.0.0"
+      (Error (`Msg "Ipaddr: lowest address has been reached")) (pred ip1);
+    ()
+
+  let test_prefix_first_last () =
+    let open V4.Prefix in
+    let assert_equal = assert_equal ~printer:V4.to_string in
+    assert_equal ~msg:"first 192.168.1.0/24"
+      (V4.of_string_exn "192.168.1.1")
+      (first (of_string_exn "192.168.1.0/24"));
+    assert_equal ~msg:"first 169.254.169.254/31"
+      (Ipaddr.V4.of_string_exn "169.254.169.254")
+      (first (of_string_exn "169.254.169.254/31"));
+    assert_equal ~msg:"first 169.254.169.254/32"
+      (Ipaddr.V4.of_string_exn "169.254.169.254")
+      (first (of_string_exn "169.254.169.254/32"));
+    assert_equal ~msg:"last 192.168.1.0/24"
+      (Ipaddr.V4.of_string_exn "192.168.1.254")
+      (last (of_string_exn "192.168.1.0/24"));
+    assert_equal ~msg:"last 169.254.169.254/31"
+      (Ipaddr.V4.of_string_exn "169.254.169.255")
+      (last (of_string_exn "169.254.169.254/31"));
+    assert_equal ~msg:"last 169.254.169.254/32"
+      (Ipaddr.V4.of_string_exn "169.254.169.254")
+      (last (of_string_exn "169.254.169.254/32"))
+
   let suite = "Test V4" >::: [
     "string_rt"            >:: test_string_rt;
     "string_rt_bad"        >:: test_string_rt_bad;
@@ -327,6 +399,9 @@ module Test_v4 = struct
     "special_addr"         >:: test_special_addr;
     "multicast_mac"        >:: test_multicast_mac;
     "domain_name"          >:: test_domain_name;
+    "prefix_mem"           >:: test_prefix_mem;
+    "succ_pred"            >:: test_succ_pred;
+    "prefix_first_last"    >:: test_prefix_first_last;
   ]
 end
 
@@ -466,7 +541,7 @@ module Test_v6 = struct
       "[::]/64",                  "::/64";
     ] in
     List.iter (fun (subnet,rt) ->
-      let os = V6.Prefix.of_string_exn subnet in
+      let os = V6.Prefix.of_string_exn subnet |> V6.Prefix.prefix in
       let ts = V6.Prefix.to_string os in
       assert_equal ~msg:subnet ts rt;
       let os = Ipaddr_sexp.(V6.Prefix.(t_of_sexp (sexp_of_t os))) in
@@ -494,10 +569,14 @@ module Test_v6 = struct
     List.iter (fun (netaddr,net,addr) ->
       let netv4 = V6.Prefix.of_string_exn net in
       let addrv4 = V6.of_string_exn addr in
-      let prefix,v4 = V6.Prefix.of_address_string_exn netaddr in
+      let cidr = V6.Prefix.of_string_exn netaddr in
+      let prefix = V6.Prefix.prefix cidr
+      and v4 = V6.Prefix.address cidr
+      in
+      let prefix = V6.Prefix.prefix prefix in
       assert_equal ~msg:(net^" <> "^(V6.Prefix.to_string prefix)) netv4 prefix;
       assert_equal ~msg:(addr^" <> "^(V6.to_string v4)) addrv4 v4;
-      let addrstr = V6.Prefix.to_address_string prefix v4 in
+      let addrstr = V6.Prefix.to_string cidr in
       assert_equal ~msg:(netaddr^" <> "^addrstr) netaddr addrstr;
     ) netaddrs
 
@@ -527,13 +606,16 @@ module Test_v6 = struct
       "8::1/0",  "::";
     ] in
     List.iter (fun (net_str,nm_str) ->
-      let prefix, v6 = V6.Prefix.of_address_string_exn net_str in
-      let nm = V6.Prefix.netmask prefix in
-      let nnm_str = V6.to_string nm in
+      let cidr = V6.Prefix.of_string_exn net_str in
+      let prefix = V6.Prefix.prefix cidr
+      and address = V6.Prefix.address cidr
+      in
+      let netmask = V6.Prefix.netmask prefix in
+      let nnm_str = V6.to_string netmask in
       let msg = Printf.sprintf "netmask %s <> %s" nnm_str nm_str in
       assert_equal ~msg nnm_str nm_str;
-      let prefix = V6.Prefix.of_netmask nm v6 in
-      let nns = V6.Prefix.to_address_string prefix v6 in
+      let prefix = V6.Prefix.of_netmask_exn ~netmask ~address in
+      let nns = V6.Prefix.to_string prefix in
       let msg = Printf.sprintf "%s is %s under netmask iso" net_str nns in
       assert_equal ~msg net_str nns
     ) nets
@@ -546,9 +628,10 @@ module Test_v6 = struct
       error "ffff:fffe:8000::" "invalid netmask";
     ] in
     List.iter (fun (nm_str,exn) ->
-      let nm = V6.of_string_exn nm_str in
-      let addr = V6.of_string_exn "::" in
-      assert_raises ~msg:nm_str exn (fun () -> V6.Prefix.of_netmask nm addr)
+      let netmask = V6.of_string_exn nm_str in
+      let address = V6.of_string_exn "::" in
+      assert_raises ~msg:nm_str exn
+        (fun () -> V6.Prefix.of_netmask_exn ~netmask ~address)
     ) bad_masks
 
   let test_scope () =
@@ -599,23 +682,19 @@ module Test_v6 = struct
 
   let test_prefix_map () =
     let module M = Map.Make(V6.Prefix) in
-    let m = M.add (V6.Prefix.of_string_exn "::ffff:0.0.0.0/0")
-      "everyone" M.empty in
-    let m = M.add (V6.Prefix.of_string_exn "::ffff:192.0.0.0/1")
-      "weirdos" m in
-    let m = M.add (V6.Prefix.of_string_exn "::ffff:128.0.0.0/1")
-      "high-bitters" m in
-    let m = M.add (V6.Prefix.of_string_exn "::ffff:254.0.0.0/8")
-      "top-end" m in
-    let m = M.add (V6.Prefix.of_string_exn "::ffff:0.0.0.0/0")
-      "iana" m in
+    let of_string s = s |> V6.Prefix.of_string_exn |> V6.Prefix.prefix in
+    let m = M.add (of_string "::ffff:0.0.0.0/0") "everyone" M.empty in
+    let m = M.add (of_string "::ffff:192.0.0.0/1") "weirdos" m in
+    let m = M.add (of_string "::ffff:128.0.0.0/1") "high-bitters" m in
+    let m = M.add (of_string "::ffff:254.0.0.0/8") "top-end" m in
+    let m = M.add (of_string "::ffff:0.0.0.0/0") "iana" m in
     assert_equal ~msg:"size" (M.cardinal m) 3;
     assert_equal ~msg:"min" (M.min_binding m)
-      (V6.Prefix.of_string_exn "::ffff:0.0.0.0/0", "iana");
+      (of_string "::ffff:0.0.0.0/0", "iana");
     assert_equal ~msg:"max" (M.max_binding m)
-      (V6.Prefix.of_string_exn "::ffff:254.0.0.0/8", "top-end");
+      (of_string "::ffff:254.0.0.0/8", "top-end");
     assert_equal ~msg:"third"
-      (M.find (V6.Prefix.of_string_exn "::ffff:128.0.0.0/1") m) "high-bitters"
+      (M.find (of_string "::ffff:128.0.0.0/1") m) "high-bitters"
 
   let test_multicast_mac () =
     let on = 0xFFFF in
@@ -647,6 +726,55 @@ module Test_v6 = struct
     assert_equal ~msg:("link_address_of_mac "^ip_str^" <> "^expected)
       ip_str expected
 
+  let test_succ_pred () =
+    let open V6 in
+    let printer = function
+      | Ok v -> Printf.sprintf "Ok %s" (V6.to_string v)
+      | Error (`Msg e) -> Printf.sprintf "Error `Msg \"%s\"" e
+    in
+    let assert_equal = assert_equal ~printer in
+    let ip1 = of_string_exn "::" in
+    let ip2 = of_string_exn "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff" in
+    let ip3 = of_string_exn "::2" in
+    assert_equal ~msg:"succ ::" (of_string "::1") (succ ip1);
+    assert_equal ~msg:"succ (succ ::)"
+      (of_string "::2") (succ ip1 >>= succ);
+    assert_equal ~msg:"succ ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff"
+      (Error (`Msg "Ipaddr: highest address has been reached")) (succ ip2);
+    assert_equal ~msg:"pred ::2" (of_string "::1") (pred ip3) ;
+    assert_equal ~msg:"pred ::ffff:ffff"
+      (of_string "::ffff:fffd")
+      (of_string "::ffff:ffff" >>= pred >>= pred);
+    assert_equal ~msg:"pred ::"
+      (Error (`Msg "Ipaddr: lowest address has been reached")) (pred ip1);
+    assert_equal ~msg:"pred (succ ::2)" (Ok ip3) (succ ip3 >>= pred)
+
+  let test_first_last () =
+    let open V6 in
+    let open Prefix in
+    let ip_of_string = V6.of_string_exn in
+    let assert_equal = assert_equal ~printer:V6.to_string in
+    assert_equal ~msg:"first ::/64"
+      (ip_of_string "::1") (first @@ of_string_exn "::/64");
+    assert_equal ~msg:"first ::ff00/120"
+      (ip_of_string "::ff01") (first @@ of_string_exn "::ff00/120");
+    assert_equal ~msg:"first ::aaa0/127"
+      (ip_of_string "::aaa0") (first @@ of_string_exn "::aaa0/127");
+    assert_equal ~msg:"first ::aaa0/128" (ip_of_string "::aaa0")
+      (first @@ of_string_exn "::aaa0/128");
+    assert_equal ~msg:"last ::/64" (ip_of_string "::ffff:ffff:ffff:ffff")
+      (last @@ of_string_exn "::/64");
+    assert_equal ~msg:"last ::/120" (ip_of_string "::ff")
+      (last @@ of_string_exn "::/120");
+    assert_equal ~msg:"last ::/112" (ip_of_string "::ffff")
+      (last @@ of_string_exn "::/112");
+    assert_equal ~msg:"last ::bbbb:eeee:0000:0000/64" (ip_of_string "::ffff:ffff:ffff:ffff")
+      (last @@ of_string_exn "::bbbb:eeee:0000:0000/64");
+    assert_equal ~msg:"last ::aaa0/127" (ip_of_string "::aaa1")
+      (last @@ of_string_exn "::aaa0/127");
+    assert_equal ~msg:"last ::aaa0/128" (ip_of_string "::aaa0")
+      (last @@ of_string_exn "::aaa0/128")
+
   let suite = "Test V6" >::: [
     "string_rt"            >:: test_string_rt;
     "string_rt_bad"        >:: test_string_rt_bad;
@@ -669,6 +797,8 @@ module Test_v6 = struct
     "multicast_mac"        >:: test_multicast_mac;
     "domain_name"          >:: test_domain_name;
     "link_address_of_mac"  >:: test_link_address_of_mac;
+    "succ_pred"            >:: test_succ_pred;
+    "first_last"           >:: test_first_last;
   ]
 end
 
